from abc import ABC, abstractmethod

from django.core.exceptions import ValidationError


class BaseValidator(ABC):
    message = "Invalid value"
    code = "invalid"

    @staticmethod
    @abstractmethod
    def validation(value: any) -> callable:
        raise NotImplemented

    def __call__(self, value: str):
        if not self.validation(value):
            raise ValidationError(message=self.message, code=self.code)



from django_utk.utils.args import EMPTY


def popattr(obj: object, attr_name: str, default: any = EMPTY):
    try:
        attr, _ = getattr(obj, attr_name), delattr(obj, attr_name)
    except AttributeError as exc:
        if default is EMPTY:
            raise exc
        attr = default

    return attr



from unittest import TestCase
from unittest.mock import MagicMock

from django_utk.tests import faker
from django_utk.validators.uuid import is_valid_uuid

rand_int = faker.RandInt(1, 10)


class RandGeneratorTestCase(TestCase):
    def test__default(self):
        length = rand_int()
        generator_fct = faker.RandGenerator(length)
        generator = generator_fct()

        self.assertEqual(len([i for i in generator]), length)

        with self.assertRaises(StopIteration):
            next(generator)

    def test__item_factory__callable(self):
        length = rand_int()
        items = []

        def item_factory():
            item = MagicMock()
            items.append(item)
            return item

        generator_fct = faker.RandGenerator(length, item_factory=item_factory)
        generator = generator_fct()
        generated_list = [item for item in generator]

        self.assertListEqual(items, generated_list)

        with self.assertRaises(StopIteration):
            next(generator)

    def test__item_factory__factory(self):
        length = rand_int()
        item_factory = faker.UUID4()

        generator_fct = faker.RandGenerator(length, item_factory=item_factory)

        for item in generator_fct():
            self.assertTrue(is_valid_uuid(item))


class RandListTestCase(TestCase):
    def test__default(self):
        length = rand_int()
        list_fct = faker.RandList(length)
        the_list = list_fct()

        self.assertIsInstance(the_list, list)
        self.assertEqual(len(the_list), length)


class RandTupleTestCase(TestCase):
    def test__default(self):
        length = rand_int()
        tuple_fct = faker.RandTuple(length)
        the_tuple = tuple_fct()

        self.assertIsInstance(the_tuple, tuple)
        self.assertEqual(len(the_tuple), length)



from django_utk.db.models.dbview.managers import *
from django_utk.db.models.dbview.managers import __all__ as managers_all
from django_utk.db.models.dbview.mixins import *
from django_utk.db.models.dbview.mixins import __all__ as mixins_all
from django_utk.db.models.dbview.models import *
from django_utk.db.models.dbview.models import __all__ as models_all
from django_utk.db.models.dbview.querysets import *
from django_utk.db.models.dbview.querysets import __all__ as querysets_all

__all__ = [
    *querysets_all,
    *managers_all,
    *mixins_all,
    *models_all,
]



import random
from typing import Any, Iterable

from django_utk.tests.faker.base import DataFactory

__all__ = [
    "RandChoices",
    "RandChoice",
]

from django_utk.utils.typehint import typehint


class RandChoices(DataFactory):
    getter = random.choices

    def __init__(
        self,
        population: Iterable[Any],
        k=1,
        *,
        weights: list[int] = None,
        cum_weights: list[int] = None,
    ):
        super().__init__(
            population=population,
            weights=weights,
            cum_weights=cum_weights,
            k=k,
        )

    @typehint(DataFactory)
    def __call__(
        self,
        *,
        population: Iterable[Any],
        k: int = 1,
        weights: list[int] = None,
        cum_weights: list[int] = None,
    ) -> list:
        pass


class RandChoice(RandChoices):
    def __init__(
        self,
        population: Iterable[Any],
        *,
        weights: list[int] = None,
        cum_weights: list[int] = None,
    ):
        super().__init__(
            population=population,
            weights=weights,
            cum_weights=cum_weights,
            k=1,
        )

    def getter(self, *args, **kwargs):
        try:
            return super().getter(*args, **kwargs)[0]
        except IndexError:
            return None

    @typehint(RandChoices)
    def __call__(
        self,
        *,
        population: Iterable[Any] = None,
        k: int = 1,
        weights: list[int] = None,
        cum_weights: list[int] = None,
    ):
        pass



import pathlib
import random
import string
from typing import Callable

from django_utk.tests.faker.base import DataFactory
from django_utk.utils.typehint import typehint

__all__ = [
    "RandString",
    "RandFilename",
    "RandPath",
    "RandFilePath",
]


class RandString(DataFactory):
    ALPHABET = string.hexdigits
    LENGTH = 32

    @staticmethod
    def getter(length: int | Callable, prefix: str, suffix: str, alphabet: str):
        if callable(length):
            length = length()

        return "{prefix}{string}{suffix}".format(
            prefix=prefix,
            string="".join(random.choices(alphabet, k=length)),
            suffix=suffix,
        )

    def __init__(
        self,
        length: int | Callable = LENGTH,
        *,
        prefix: str = None,
        suffix: str = None,
        alphabet: str = None,
        **kwargs,
    ):
        super().__init__(
            length=length,
            prefix=prefix or "",
            suffix=suffix or "",
            alphabet=alphabet or self.ALPHABET,
        )

    @typehint(DataFactory)
    def __call__(
        self,
        *,
        length: int | Callable = None,
        prefix: str = None,
        suffix: str = None,
        alphabet: str = None,
    ) -> str:
        pass


class RandFilename(RandString):
    LENGTH = 64
    ALPHABET = string.ascii_letters
    EXTENSIONS = [
        *["3dm", "3ds"],
        *["aac", "accdb", "ai", "aiff", "apk", "app", "asp", "aspx", "avi"],
        *["bat", "bin", "bmp"],
        *["c", "cer", "class", "cmd", "com", "conf", "cpp", "crt", "csv"],
        *["dat", "db", "dbf", "dll", "doc", "docx", "dmg", "dmp", "dxf"],
        *["eps", "exe"],
        *["fla", "flac", "flv"],
        *["gif", "gz", "gzip"],
        *["h", "htm", "html", "hqx", "htaccess"],
        *["ics", "iff", "inc", "ini", "iso", "iso"],
        *["jar", "java", "jpg", "js", "json", "jsp"],
        *["key"],
        *["log"],
        *["mov", "mp2", "mp3", "mp4", "mpeg", "mpg", "msg", "msi", "mswmm"],
        *["obj", "odp", "ods", "odt", "otp", "ots", "ott"],
        *["pdf", "pkg", "png", "pptx", "ps", "psd", "pub"],
        *["rar", "rb", "reg", "rpm", "rss", "rtf"],
        *["sav", "sh", "sitx", "sql", "svg", "swf", "sys"],
        *["tar", "tif", "tiff", "tmp", "torrent", "ttf", "txt"],
        *["v", "vcf", "vob"],
        *["wav", "webm", "wps", "xlr", "xls", "xlsx", "xml"],
        *["z", "zip", "7z"],
    ]

    def getter(self, *args, **kwargs):
        return "{name}.{extension}".format(
            name=super().getter(*args, **kwargs),
            extension=random.choices(self.extensions)[0],
        )

    def __init__(
        self,
        length: int = LENGTH,
        *,
        extensions: list[str] = None,
        alphabet: str = None,
    ):
        self.extensions = extensions or self.EXTENSIONS

        super().__init__(length=length, alphabet=alphabet)

    @typehint(RandString)
    def __call__(
        self,
        *,
        length: int | Callable = None,
        alphabet: str = None,
    ) -> str:
        pass


class RandPath(RandString):
    ALPHABET = string.ascii_letters

    def getter(self, *args, **kwargs):
        if callable(self.depth):
            depth = self.depth()
        else:
            depth = self.depth

        directories = [
            super(RandPath, self).getter(*args, **kwargs) for _ in range(depth)
        ]
        path = pathlib.Path().joinpath(self.root, *directories)

        return f"{path}{self.end}"

    def __init__(self, root: str = None, depth: int | Callable = None, end: str = "/"):
        from django_utk.tests.faker import RandInt

        self.root = root or ""
        self.depth = depth or RandInt(0, 5)
        self.end = end

        super().__init__(length=RandInt(1, 10), alphabet=self.ALPHABET)

    @typehint(RandString)
    def __call__(self) -> str:
        pass


class RandFilePath(RandPath):

    def getter(self, *args, **kwargs):
        path = super().getter(*args, **kwargs)
        file = self.filename_factory()
        return f"{path}{file}"

    def __init__(
        self,
        root: str = None,
        depth: int | Callable = None,
    ):
        self.filename_factory = RandFilename()

        super().__init__(root=root, depth=depth)

    @typehint(RandPath)
    def __call__(self) -> str:
        pass



from abc import ABC, abstractmethod
from typing import Any, Callable, Generator, Iterable, TypeVar

from django_utk.tests.faker.base import DataFactory
from django_utk.utils.typehint import typehint

__all__ = [
    "Sequence",
    "ForEach",
]


T = TypeVar("T")


class BaseSequence(DataFactory):
    def __init__(self, *args, **kwargs):
        self.sequencer: Generator[T, None, None] = self.get_sequencer()
        super().__init__()

    def __call__(self):
        return super().__call__()

    def getter(self):
        return next(self.sequencer)

    @abstractmethod
    def get_sequencer(self) -> Generator[Any, None, None]:
        raise NotImplemented


class Sequence(BaseSequence):
    @staticmethod
    def default_handler(x: int) -> int:
        return x

    def __init__(
        self,
        handler: Callable[[int], T] = None,
        start: int = 0,
        step: int = 1,
        end: int = None,
    ):
        self.handler = handler or self.default_handler
        self.start = start
        self.step = step
        self.end = end
        self.current = start

        super().__init__()

    def get_sequencer(self) -> Generator[Any, None, None]:
        if self.end is not None:
            for i in range(self.start, self.end, self.step):
                self.current = i
                yield self.handler(self.current)
        else:
            while True:
                yield self.handler(self.current)
                self.current += self.step

    @typehint(BaseSequence)
    def __call__(self) -> any:
        pass


class ForEach(BaseSequence):
    def __init__(self, items: Iterable[T]):
        self.items = items
        super(ForEach, self).__init__()

    def get_sequencer(self) -> Generator[Any, None, None]:
        for item in self.items:
            yield item

    @typehint(BaseSequence)
    def __call__(self) -> any:
        pass



import io
import random
from typing import Callable

import django.core.files
import django.core.files.images

from django_utk.tests.faker.base import DataFactory, TypeCasting
from django_utk.tests.faker.numeric import RandInt
from django_utk.tests.faker.strings import RandFilename
from django_utk.utils.typehint import typehint

__all__ = [
    "RandBytes",
    "RandBytesIO",
    "RandFile",
    "RandImageFile",
]


class RandBytes(DataFactory):
    def getter(self, n: int | DataFactory | Callable):
        if callable(n):
            n = n()
        return random.randbytes(n)

    def __init__(self, n: int | DataFactory | Callable = None, **kwargs):
        if n is None:
            n = RandInt(0, (kilobyte := 1000))

        super().__init__(n=n, **kwargs)

    @typehint(DataFactory)
    def __call__(self, /, n: int | DataFactory | Callable = None) -> bytes:
        pass


class RandBytesIO(TypeCasting, RandBytes):
    cast = io.BytesIO

    @typehint(RandBytes)
    def __call__(self, /, n: int | DataFactory | Callable) -> io.BytesIO:
        pass


class RandFile(RandBytesIO):
    local_cast = django.core.files.File

    def getter(
        self, n: int | DataFactory | Callable, name: str | DataFactory | Callable
    ):
        if callable(name):
            name = name()
        return self.local_cast(super().getter(n), name=name)

    def __init__(
        self,
        n: int | DataFactory | Callable = None,
        name: str | DataFactory | Callable = None,
    ):
        if name is None:
            name = RandFilename()

        super().__init__(n=n, name=name)

    @typehint(RandBytesIO)
    def __call__(
        self,
        /,
        n: int | DataFactory | Callable,
        name: str | DataFactory | Callable,
    ) -> django.core.files.File:
        pass


class RandImageFile(RandFile):
    local_cast = django.core.files.images.ImageFile
    IMAGE_EXTENSIONS = [
        "jpg",
        "jpeg",
        "png",
        "gif",
        "bmp",
        "tif",
        "tiff",
        "webp",
        "svg",
        "ico",
        "heic",
    ]

    def __init__(
        self,
        n: int | DataFactory | Callable = None,
        name: str | DataFactory | Callable = None,
    ):
        if name is None:
            name = RandFilename(extensions=self.IMAGE_EXTENSIONS)

        super().__init__(n=n, name=name)

    @typehint(RandBytesIO)
    def __call__(
        self,
        *,
        n: int | DataFactory | Callable,
        name: str | DataFactory | Callable,
    ) -> django.core.files.images.ImageFile:
        pass



from django.db import models
from django.utils.translation import gettext_lazy as _

from django_utk.db.models.timestamped.fields import CreatedAtField, UpdatedAtField
from django_utk.db.models.timestamped.managers import TimeStampedManager

__all__ = [
    "TimeStampedModel",
]


class TimeStampedModel(models.Model):
    """
    A TimeStampedModel model automatically records the time when an object
    is created (`created_at`) and updated (`updated_at`). This is useful
    for keeping track of when records are added or modified in the
    database, providing crucial information for auditing, analytics,
    and debugging purposes.

    Attributes:
        created_at (CreatedAtField): DateTimeField indicating the timestamp when
            the object was created.
        updated_at (UpdatedAtField): DateTimeField indicating the timestamp of
            the last update to the object.

    Manager:
        objects (TimeStampedManager): Automatically updates the `created_at`
            and `updated_at` fields on object creation and modification.
    """

    created_at = CreatedAtField(_("Created at"))
    updated_at = UpdatedAtField(_("Updated at"))

    objects = TimeStampedManager()

    class Meta:
        abstract = True



from django_utk.db.models.fix.delete_using_manager import *
from django_utk.db.models.fix.delete_using_manager import (
    __all__ as delete_using_manager_all,
)

__all__ = [*delete_using_manager_all]



from unittest import TestCase
from unittest.mock import MagicMock

from django_utk.utils.typehint import typehint


class MockType:
    def __new__(cls) -> MagicMock:
        return MagicMock(spec=type)


class TypehintTestCase(TestCase):

    def assertAnnotations(self, method: any, annotations: dict):
        method_annotations = method.__annotations__

        self.assertListEqual(list(method_annotations.keys()), list(annotations.keys()))

        for annotation_name, annotation in annotations.items():
            self.assertIs(method_annotations[annotation_name], annotation)

    def assertClassAnnotations(self, klass: type, annotations: dict):
        for method_name, method_annotations in annotations.items():
            self.assertAnnotations(getattr(klass, method_name), method_annotations)

    def test__methods(self):
        method_a__return = MockType()
        method_b__return = MockType()
        method_c__return = MockType()
        method_d__return = MockType()
        method_e__return = MockType()
        method_f__return = MockType()

        arg_first__type = MockType()
        arg_second__type = MockType()
        arg_third__type = MockType()

        class ClassA:
            call_handler = MagicMock()
            call_handler.return_value = MagicMock()

            def method_a(self):
                return self.call_handler()

            def method_b(self, first):
                return self.call_handler(first=first)

            def method_c(self, first, second):
                return self.call_handler(first=first, second=second)

            def method_d(self, first, second, *, third):
                return self.call_handler(first=first, second=second, third=third)

            def method_e(self, *first, **second):
                return self.call_handler(first=first, second=second)

            @classmethod
            def method_f(cls):
                return cls.call_handler()

        class ClassB(ClassA):
            @typehint
            def method_a(
                self,
            ) -> method_a__return:
                pass

            @typehint
            def method_b(
                self,
                first: arg_first__type,
            ) -> method_b__return:
                pass

            @typehint
            def method_c(
                self,
                first: arg_first__type,
                second: arg_second__type,
            ) -> method_c__return:
                pass

            @typehint
            def method_d(
                self,
                first: arg_first__type,
                second: arg_second__type,
                *,
                third: arg_third__type,
            ) -> method_d__return:
                pass

            @typehint
            def method_e(
                self,
                *first: arg_first__type,
                **second: arg_second__type,
            ) -> method_e__return:
                pass

            @classmethod
            @typehint
            def method_f(
                cls,
            ) -> method_f__return:
                return cls.call_handler()

        class ClassC(ClassA):
            @typehint(ClassA)
            def method_a(
                self,
            ) -> method_a__return:
                pass

            @typehint(ClassA)
            def method_b(
                self,
                first: arg_first__type,
            ) -> method_b__return:
                pass

            @typehint(ClassA)
            def method_c(
                self,
                first: arg_first__type,
                second: arg_second__type,
            ) -> method_c__return:
                pass

            @typehint(ClassA)
            def method_d(
                self,
                first: arg_first__type,
                second: arg_second__type,
                *,
                third: arg_third__type,
            ) -> method_d__return:
                pass

            @typehint(ClassA)
            def method_e(
                self,
                *first: arg_first__type,
                **second: arg_second__type,
            ) -> method_e__return:
                pass

            @classmethod
            @typehint(ClassA)
            def method_f(
                cls,
            ) -> method_f__return:
                return cls.call_handler()

        annotations_mapping = {
            "method_a": {
                "return": method_a__return,
            },
            "method_b": {
                "first": arg_first__type,
                "return": method_b__return,
            },
            "method_c": {
                "first": arg_first__type,
                "second": arg_second__type,
                "return": method_c__return,
            },
            "method_d": {
                "first": arg_first__type,
                "second": arg_second__type,
                "third": arg_third__type,
                "return": method_d__return,
            },
            "method_e": {
                "first": arg_first__type,
                "second": arg_second__type,
                "return": method_e__return,
            },
            "method_f": {
                "return": method_f__return,
            },
        }

        self.assertClassAnnotations(klass=ClassB, annotations=annotations_mapping)
        self.assertClassAnnotations(klass=ClassC, annotations=annotations_mapping)

        original = ClassA()
        instance_b = ClassB()
        instance_c = ClassC()

        _first = MagicMock()
        _second = MagicMock()
        _third = MagicMock()

        instance_c.method_f()

        for instance in [instance_b, instance_c]:
            # method A
            self.assertIs(original.method_a(), instance.method_a())

            # method B
            self.assertIs(original.method_b(_first), instance.method_b(_first))
            self.assertIs(original.call_handler.call_args.kwargs["first"], _first)

            self.assertIs(
                original.method_b(first=_first),
                instance.method_b(first=_first),
            )
            self.assertIs(original.call_handler.call_args.kwargs["first"], _first)

            # method C
            self.assertIs(
                original.method_c(_first, _second),
                instance.method_c(_first, _second),
            )
            self.assertIs(original.call_handler.call_args.kwargs["first"], _first)
            self.assertIs(original.call_handler.call_args.kwargs["second"], _second)

            self.assertIs(
                original.method_c(_first, second=_second),
                instance.method_c(_first, second=_second),
            )
            self.assertIs(original.call_handler.call_args.kwargs["first"], _first)
            self.assertIs(original.call_handler.call_args.kwargs["second"], _second)

            self.assertIs(
                original.method_c(first=_first, second=_second),
                instance.method_c(first=_first, second=_second),
            )
            self.assertIs(original.call_handler.call_args.kwargs["first"], _first)
            self.assertIs(original.call_handler.call_args.kwargs["second"], _second)

            # method D
            self.assertIs(
                original.method_d(_first, _second, third=_third),
                instance.method_d(_first, _second, third=_third),
            )
            self.assertIs(original.call_handler.call_args.kwargs["first"], _first)
            self.assertIs(original.call_handler.call_args.kwargs["second"], _second)
            self.assertIs(original.call_handler.call_args.kwargs["third"], _third)

            self.assertIs(
                original.method_d(_first, second=_second, third=_third),
                instance.method_d(_first, second=_second, third=_third),
            )
            self.assertIs(original.call_handler.call_args.kwargs["first"], _first)
            self.assertIs(original.call_handler.call_args.kwargs["second"], _second)
            self.assertIs(original.call_handler.call_args.kwargs["third"], _third)

            self.assertIs(
                original.method_d(first=_first, second=_second, third=_third),
                instance.method_d(first=_first, second=_second, third=_third),
            )
            self.assertIs(original.call_handler.call_args.kwargs["first"], _first)
            self.assertIs(original.call_handler.call_args.kwargs["second"], _second)
            self.assertIs(original.call_handler.call_args.kwargs["third"], _third)

            # method E
            self.assertIs(
                original.method_e(_first, second=_second),
                instance.method_e(_first, second=_second),
            )
            self.assertTupleEqual(
                original.call_handler.call_args.kwargs["first"],
                (_first,),
            )
            self.assertDictEqual(
                original.call_handler.call_args.kwargs["second"],
                {"second": _second},
            )

            self.assertIs(
                original.method_e(_first, _first, second_1=_second, second_2=_second),
                instance.method_e(_first, _first, second_1=_second, second_2=_second),
            )
            self.assertTupleEqual(
                original.call_handler.call_args.kwargs["first"],
                (_first, _first),
            )
            self.assertDictEqual(
                original.call_handler.call_args.kwargs["second"],
                {"second_1": _second, "second_2": _second},
            )

            # method F
            self.assertIs(original.method_f(), instance.method_f())

    def test__magic(self):
        class ClassA:
            call_handler = MagicMock()
            call_handler__return = MagicMock()
            call_handler.return_value = call_handler__return

            def __new__(cls, *args, **kwargs):
                cls.call_handler(*args, **kwargs)
                return super().__new__(cls)

            def __init__(self, *args, **kwargs):
                self.call_handler(*args, **kwargs)

            def __call__(self, *args, **kwargs):
                return self.call_handler(*args, **kwargs)

        class ClassB(ClassA):
            @typehint
            def __new__(cls, *args, **kwargs):
                pass

            @typehint
            def __init__(self, *args, **kwargs):
                pass

            @typehint
            def __call__(self, *args, **kwargs):
                pass

        class ClassC(ClassA):
            @typehint(ClassA)
            def __new__(cls, *args, **kwargs):
                pass

            @typehint(ClassA)
            def __init__(self, *args, **kwargs):
                pass

            @typehint(ClassA)
            def __call__(self, *args, **kwargs):
                pass

        for Klass in [ClassB, ClassC]:
            data = MagicMock()
            instance = Klass(data=data)

            self.assertIsInstance(instance, Klass)
            self.assertIsInstance(instance, ClassA)

            for args in ClassA.call_handler.call_args_list[-2:]:
                # last two calls: init and new
                self.assertIs(args.kwargs["data"], data)
                self.assertIs(args.kwargs["data"], data)

            self.assertIs(instance(), ClassA.call_handler__return)



from abc import ABC, abstractmethod
from typing import Any, Callable, TypeVar

from django_utk.utils.typehint import typehint

T = TypeVar("T")
CastType = T | Callable[[Any], T]


class DataFactory(ABC):

    @abstractmethod
    def getter(self, *args, **kwargs):
        raise NotImplemented

    @classmethod
    def get(cls, *args, **kwargs):
        return cls(*args, **kwargs)()

    def __call__(self, **kwargs):
        kwargs = {**self.kwargs, **kwargs}
        return self.getter(*self.args, **kwargs)

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class DataFactoryMixin(ABC):
    pass


class TypeCasting(DataFactoryMixin):
    cast: CastType = lambda x: x

    def getter(self, *args, **kwargs):
        return self.cast(super().getter(*args, **kwargs))

    @typehint(DataFactory)
    def __call__(self, *args, **kwargs) -> CastType:
        pass



from common.abc import ABCManager
from django.db import models

from django_utk.db.models.softdelete.querysets import SoftDeleteQuerySet

__all__ = [
    "BaseSoftDeleteManager",
    "SoftDeleteManager",
]


class BaseSoftDeleteManager(ABCManager):
    QuerySetClass = SoftDeleteQuerySet

    def get_queryset(self):
        return super().get_queryset().filter(deleted_at__isnull=True)


class SoftDeleteManager(BaseSoftDeleteManager, models.Manager):
    pass



from typing import Callable

from django_utk.tests.faker import RandInt
from django_utk.tests.faker.base import DataFactory
from django_utk.utils.typehint import typehint

__all__ = [
    "RandRange",
]


class RandRange(DataFactory):
    DEFAULT_MAX_SIZE = 100

    def getter(self, start: int, end: int | Callable, step: int) -> range:
        if callable(end):
            end = end()
        return range(start, end, step)

    def __init__(self, start: int = 0, end: int | Callable = None, step: int = 1):
        super(RandRange, self).__init__(
            start=start,
            end=end or RandInt(start + 1, start + self.DEFAULT_MAX_SIZE),
            step=step,
        )

    @typehint(DataFactory)
    def __call__(
        self,
        *,
        start: int = None,
        end: int | Callable = None,
        step: int = None,
    ) -> range:
        pass



from django.db import models

from django_utk.db.models.dbview import DBViewQuerySet

__all__ = [
    "DBViewManager",
]

DBViewManager = models.Manager.from_queryset(DBViewQuerySet, class_name="DBViewManager")



from functools import wraps

__all__ = [
    "run_n_times",
    "run_10_times",
    "run_100_times",
    "run_1k_times",
    "run_10k_times",
    "run_100k_times",
    "run_1m_times",
]


def run_n_times(n: int):
    def repeat_wrapper(method: callable):
        @wraps(method)
        def many_running_method(*args, **kwargs):
            for _ in range(n):
                method(*args, **kwargs)

        return many_running_method

    return repeat_wrapper


run_10_times = run_n_times(10)
run_100_times = run_n_times(100)
run_1k_times = run_n_times(1_000)
run_10k_times = run_n_times(10_000)
run_100k_times = run_n_times(100_000)
run_1m_times = run_n_times(1_000_000)



from django_utk.tests.faker.byteslike import *
from django_utk.tests.faker.byteslike import __all__ as byteslike_all
from django_utk.tests.faker.choices import *
from django_utk.tests.faker.choices import __all__ as choices_all
from django_utk.tests.faker.hashes import *
from django_utk.tests.faker.hashes import __all__ as hashes_all
from django_utk.tests.faker.iterable import *
from django_utk.tests.faker.iterable import __all__ as iterable_all
from django_utk.tests.faker.numeric import *
from django_utk.tests.faker.numeric import __all__ as numeral_all
from django_utk.tests.faker.ranges import *
from django_utk.tests.faker.ranges import __all__ as ranges_all
from django_utk.tests.faker.sequences import *
from django_utk.tests.faker.sequences import __all__ as sequences_all
from django_utk.tests.faker.strings import *
from django_utk.tests.faker.strings import __all__ as choices_all
from django_utk.tests.faker.uuid import *
from django_utk.tests.faker.uuid import __all__ as uuid_all

__all__ = [
    *byteslike_all,
    *choices_all,
    *hashes_all,
    *iterable_all,
    *numeral_all,
    *ranges_all,
    *sequences_all,
    *choices_all,
    *uuid_all,
]



from django.db import connection

from django_utk.db.models import ModelMixin
from django_utk.db.models.utils import get_model_meta as meta

__all__ = [
    "MaterializedDBViewModelMixin",
]


class MaterializedDBViewModelMixin(ModelMixin):
    @classmethod
    def refresh(cls):
        with connection.cursor() as cursor:
            cursor.execute(f"REFRESH MATERIALIZED VIEW {meta(cls).db_table};")

    class Meta:
        abstract = True



from django.db import models

from django_utk.db.models import fix
from django_utk.db.models.dbview import DBViewManager
from django_utk.db.models.dbview.mixins import MaterializedDBViewModelMixin

__all__ = [
    "DBViewModel",
    "MaterializedDBViewModel",
]


class DBViewModel(fix.DeleteUsingManager, models.Model):
    objects = DBViewManager()

    class Meta:
        abstract = True
        managed = False
        db_table = NotImplemented


class MaterializedDBViewModel(MaterializedDBViewModelMixin, DBViewModel):
    objects: DBViewManager  # DBViewModel

    class Meta:
        abstract = True
        managed = False
        db_table = NotImplemented



import importlib
import types

__all__ = ["import_modules"]


def _get_submodules(root: str) -> list[str]:
    submodules = importlib.import_module(root)

    return [
        f"{root}.{child_name}"
        for child_name, child in submodules.__dict__.items()
        if isinstance(child, types.ModuleType)
    ]


def import_modules(path: str):
    """
    Same as `importlib.import_module` but support '*' as "any submodule" syntax

    Example:
        >>> import_modules("website.pages.*.script")
        >>> import_modules("website.pages.*")
    """

    if "*" not in path:
        return importlib.import_module(path)

    root, after_star = path.split("*", 1)
    root = root.rstrip(".")

    if root:
        raise ImportError(
            f"Invalid input {path=!r}. "
            "Value of 'path' parameter can not starts with '*' symbol, "
            "because it may cause cyclic import."
        )

    for child in _get_submodules(root):
        if after_star:
            child = ".".join([child, after_star])
        return import_modules(child)



import io
from unittest import TestCase

import django.core.files
import django.core.files.images

from django_utk.tests import faker
from django_utk.tests.utils import run_1k_times

big_int = faker.RandInt(1, 10000)
rand_str = faker.RandString()


class BaseBytesLikeTestCase:
    FactoryClass = NotImplemented
    expected_type = NotImplemented

    def get_size(self, item) -> int:
        raise NotImplemented

    @classmethod
    def setUpClass(cls):
        cls.byteslike_fct = cls.FactoryClass()
        super().setUpClass()

    @run_1k_times
    def test__default(self):
        self.assertIsInstance(self.byteslike_fct(), self.expected_type)

    @run_1k_times
    def test__n(self):
        n = big_int()

        bytes_fct_1 = self.FactoryClass(n=n)
        bytes_fct_2 = self.FactoryClass(n)
        bytes_fct_3 = self.FactoryClass()

        result_1 = bytes_fct_1()
        result_2 = bytes_fct_2()
        result_3 = bytes_fct_3(n=n)

        self.assertIsInstance(result_1, self.expected_type)
        self.assertIsInstance(result_2, self.expected_type)
        self.assertIsInstance(result_3, self.expected_type)

        self.assertEqual(self.get_size(result_1), n)
        self.assertEqual(self.get_size(result_2), n)
        self.assertEqual(self.get_size(result_3), n)


class RandBytesTestCase(BaseBytesLikeTestCase, TestCase):
    FactoryClass = faker.RandBytes
    expected_type = bytes

    def get_size(self, item: bytes) -> int:
        return len(item)


class RandBytesIOTestCase(BaseBytesLikeTestCase, TestCase):
    FactoryClass = faker.RandBytesIO
    expected_type = io.BytesIO

    def get_size(self, item: io.BytesIO):
        return item.getbuffer().nbytes


class RandFileTestCase(BaseBytesLikeTestCase, TestCase):
    FactoryClass = faker.RandFile
    expected_type = django.core.files.File

    def get_size(self, item: django.core.files.File):
        return item.size

    @run_1k_times
    def test__name(self):
        name = rand_str()

        bytes_fct_1 = self.FactoryClass(name=name)
        bytes_fct_3 = self.FactoryClass()

        result_1 = bytes_fct_1()
        result_3 = bytes_fct_3(name=name)

        self.assertEqual(result_1.name, name)
        self.assertEqual(result_3.name, name)


class RandImageFileTestCase(RandFileTestCase, TestCase):
    FactoryClass = faker.RandImageFile
    expected_type = django.core.files.images.ImageFile



import hashlib
import random
from unittest import TestCase
from unittest.mock import MagicMock

from django.core.exceptions import ValidationError

from django_utk.tests import faker
from django_utk.tests.utils import run_1k_times
from django_utk.validators.hashes import (
    HashValidator,
    MD5Validator,
    SHA1Validator,
    SHA256Validator,
    SHA512Validator,
    is_hash_valid,
    is_valid_md5,
    is_valid_sha1,
    is_valid_sha256,
    is_valid_sha512,
)

rand_int = faker.RandInt(1, 100)
big_int = faker.RandInt(1, int(faker.RandFloat.MAX))
not_hash = faker.RandString(length=rand_int(), alphabet="qwrtyuiopsghjklzxvnm")


class HashValidationTestCase(TestCase):

    @run_1k_times
    def assertValid(self, algo: callable, validation: callable):
        hash_sum = algo(random.randbytes(10)).hexdigest()
        self.assertTrue(validation(hash_sum))

    @run_1k_times
    def assertInvalid(self, validation: callable):
        self.assertFalse(validation(not_hash()))

    def test__is_valid_md5(self):
        self.assertValid(algo=hashlib.md5, validation=is_valid_md5)
        self.assertInvalid(validation=is_valid_md5)

    def test__is_valid_sha1(self):
        self.assertValid(algo=hashlib.sha1, validation=is_valid_sha1)
        self.assertInvalid(validation=is_valid_sha1)

    def test__is_valid_sha256(self):
        self.assertValid(algo=hashlib.sha256, validation=is_valid_sha256)
        self.assertInvalid(validation=is_valid_sha256)

    def test__is_valid_sha512(self):
        self.assertValid(algo=hashlib.sha512, validation=is_valid_sha512)
        self.assertInvalid(validation=is_valid_sha512)

    def test__is_hash_valid(self):
        hash_sum = str(hex(big_int()))
        self.assertTrue(is_hash_valid(value=hash_sum, length=len(hash_sum)))

    def test__is_hash_valid__exception(self):
        for ExceptionType in (
            ValueError,
            TypeError,
            MagicMock(spec=Exception),  # some unexpected exception
        ):
            hash_sum = MagicMock()
            hash_sum.__str__ = MagicMock()
            hash_sum.__len__ = MagicMock()
            hash_sum.__str__.side_effect = ExceptionType
            hash_sum.__len__.return_value = rand_int()
            self.assertFalse(is_hash_valid(value=hash_sum, length=len(hash_sum)))


class HashValidatorsTestCase(TestCase):
    @run_1k_times
    def assertValid(self, algo: callable, validator: HashValidator):
        hash_sum = algo(random.randbytes(10)).hexdigest()
        validator(hash_sum)

    @run_1k_times
    def assertInvalid(self, validator: HashValidator):
        with self.assertRaises(ValidationError):
            validator(not_hash())

    def test__MD5Validator(self):
        self.assertValid(algo=hashlib.md5, validator=MD5Validator())
        self.assertInvalid(validator=MD5Validator())

    def test__SHA1Validator(self):
        self.assertValid(algo=hashlib.sha1, validator=SHA1Validator())
        self.assertInvalid(validator=SHA1Validator())

    def test__SHA256Validator(self):
        self.assertValid(algo=hashlib.sha256, validator=SHA256Validator())
        self.assertInvalid(validator=SHA256Validator())

    def test__SHA512Validator(self):
        self.assertValid(algo=hashlib.sha512, validator=SHA512Validator())
        self.assertInvalid(validator=SHA512Validator())



import uuid
from unittest import TestCase

from django_utk.tests import faker
from django_utk.tests.utils import run_1k_times
from django_utk.validators.uuid import is_valid_uuid1, is_valid_uuid4

uuid_length = 36
rand_int = faker.RandInt(1, (uuid_length - 1))
not_uuid = faker.RandString(length=rand_int())


class UUIDValidationTestCase(TestCase):

    @run_1k_times
    def assertValid(self, algo: callable, validation: callable):
        self.assertTrue(validation(value := algo()), f"{value=}")

    @run_1k_times
    def assertInvalid(self, validation: callable):
        self.assertFalse(validation(not_uuid()))

    def test__is_valid_uuid1(self):
        self.assertValid(algo=uuid.uuid1, validation=is_valid_uuid1)
        self.assertInvalid(validation=is_valid_uuid1)

    def test__is_valid_uuid4(self):
        self.assertValid(algo=uuid.uuid4, validation=is_valid_uuid4)
        self.assertInvalid(validation=is_valid_uuid4)



import hashlib
from abc import ABC, abstractmethod

from django_utk.tests.faker.strings import RandString
from django_utk.utils.typehint import typehint

__all__ = [
    "MD5HashSum",
    "SHA1HashSum",
    "SHA256HashSum",
    "SHA512HashSum",
]


class BaseHashSum(RandString, ABC):
    LENGTH = 1024

    @classmethod
    @property
    @abstractmethod
    def hash_algo(cls) -> callable:
        raise NotImplemented

    def getter(self, *args, **kwargs):
        return self.hash_algo(
            super().getter(*args, **kwargs).encode("utf-8")
        ).hexdigest()

    def __init__(self):
        super().__init__(
            length=self.LENGTH,
        )

    @typehint(RandString)
    def __call__(self) -> str:
        pass


class MD5HashSum(BaseHashSum):
    hash_algo = hashlib.md5


class SHA1HashSum(BaseHashSum):
    hash_algo = hashlib.sha1


class SHA256HashSum(BaseHashSum):
    hash_algo = hashlib.sha256


class SHA512HashSum(BaseHashSum):
    hash_algo = hashlib.sha512



import string
from unittest import TestCase
from unittest.mock import MagicMock

from django_utk.tests import faker
from django_utk.tests.utils.run_n_times import (
    run_1k_times,
    run_10_times,
    run_10k_times,
    run_100_times,
    run_n_times,
)

small_int = faker.RandInt(2, 42)


class RunNTimesTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.loopers = [
            (10, run_10_times),
            (100, run_100_times),
            (1_000, run_1k_times),
            (10_000, run_10k_times),
            # (100_000, run_100k_times),    # FIXME: sorry this thing is too slow
            # (1_000_000, run_1m_times),
        ] + [(n, run_n_times(n)) for n in range(small_int())]

        cls.rand_int = faker.RandInt()
        cls.rand_str = faker.RandString(alphabet=string.ascii_letters)
        cls.rand_float = faker.RandString(alphabet=string.ascii_letters)

    def test__all(self):
        def get_rand_args():
            return [self.rand_int() for _ in range(small_int())]

        def get_rand_kwargs():
            return {self.rand_str(): self.rand_float() for _ in range(small_int())}

        for n, looper in self.loopers:
            args = get_rand_args()
            kwargs = get_rand_kwargs()

            original_func = MagicMock()
            looped_func = looper(original_func)

            looped_func(*args, **kwargs)

            self.assertEqual(original_func.call_count, n)

            for i in range(n):
                self.assertListEqual(list(original_func.call_args_list[i].args), args)
                self.assertEqual(original_func.call_args_list[i].kwargs, kwargs)



from typing import Callable, Generator, Iterable, List, Tuple

from django_utk.tests.factories import Factory
from django_utk.tests.faker.base import DataFactory
from django_utk.tests.faker.numeric import RandInt
from django_utk.utils.lazy import Lazy

__all__ = [
    "RandGenerator",
    "RandIterable",
    "RandList",
    "RandTuple",
]

from django_utk.utils.typehint import typehint


class RandGenerator(DataFactory):
    def getter(self, length: int, item_factory: callable) -> Generator:
        return (item_factory() for _ in range(length))

    def __init__(
        self,
        length: int,
        item_factory: Lazy | Factory | Callable = None,
    ):
        super(RandGenerator, self).__init__(
            length=length,
            item_factory=item_factory or RandInt(),
        )

    @typehint(DataFactory)
    def __call__(self) -> Generator:
        pass


class RandIterable(RandGenerator):
    def getter(self, *args, **kwargs) -> range:
        return self.cast(super().getter(*args, **kwargs))

    def __init__(
        self,
        length: int,
        item_factory: Lazy | Factory | Callable = None,
        cast: type = None,
    ):
        self.cast = cast or (lambda x: x)

        super(RandIterable, self).__init__(length=length, item_factory=item_factory)

    @typehint(RandGenerator)
    def __call__(self) -> Iterable:
        pass


class RandList(RandIterable):
    def __init__(
        self,
        length: int,
        item_factory: Lazy | Factory | Callable = None,
    ):
        super(RandList, self).__init__(
            length=length,
            item_factory=item_factory,
            cast=list,
        )

    @typehint(RandIterable)
    def __call__(self) -> List:
        pass


class RandTuple(RandIterable):
    def __init__(
        self,
        length: int,
        item_factory: Lazy | Factory | Callable = None,
    ):
        super(RandTuple, self).__init__(
            length=length,
            item_factory=item_factory,
            cast=tuple,
        )

    @typehint(RandIterable)
    def __call__(self, *, length: int) -> Tuple:
        pass



import string
from unittest import TestCase

from django_utk.tests import faker
from django_utk.tests.utils import run_1k_times, run_100_times

rand_int = faker.RandInt(1, 42)
rand_str = faker.RandString()


class RandStringTestCase(TestCase):
    @run_100_times
    def assertResultValid(
        self,
        string_factory: callable,
        length: int = faker.RandString.LENGTH,
        alphabet: set = faker.RandString.ALPHABET,
    ):
        result = string_factory()

        self.assertEqual(len(result), length)
        self.assertEqual(len(set(result) - set(alphabet)), 0)

    def test_default(self):
        string_fct = faker.RandString()

        self.assertResultValid(string_fct)

    @run_100_times
    def test__alphabet(self):
        string_len = rand_int()
        alphabet = string.printable[: min(len(string.printable), rand_int())]
        string_fct = faker.RandString(string_len, alphabet=alphabet)
        self.assertResultValid(string_fct, string_len, set(alphabet))


class RandFilenameTestCase(TestCase):
    @run_100_times
    def assertResultValid(
        self,
        filename_factory: callable,
        extensions: list[str] = faker.RandFilename.EXTENSIONS,
        length: int = faker.RandFilename.LENGTH,
        alphabet: set = faker.RandFilename.ALPHABET,
    ):
        result: str = filename_factory()
        name, ext = result.split(".", 1)

        self.assertEqual(len(name), length)
        self.assertEqual(len(set(name) - set(alphabet)), 0)
        self.assertIn(ext, extensions)

    def test_default(self):
        string_fct = faker.RandFilename()
        self.assertResultValid(string_fct)

    @run_100_times
    def test__alphabet(self):
        string_len = rand_int()
        alphabet = string.printable[: min(len(string.printable), rand_int())]
        filename_fct = faker.RandFilename(string_len, alphabet=alphabet)
        self.assertResultValid(filename_fct, length=string_len, alphabet=set(alphabet))

    @run_100_times
    def test__extensions(self):
        rand_str = faker.RandString()
        extensions = [rand_str() for _ in range(rand_int())]
        filename_fct = faker.RandFilename(extensions=extensions)
        self.assertResultValid(filename_fct, extensions=extensions)


class RandPathTestCase(TestCase):
    @run_1k_times
    def test__default(self):
        path_fct = faker.RandPath()
        result = path_fct()

        self.assertTrue(result.endswith("/"))

        for i in range(2, len(result)):
            self.assertNotIn(("/" * i), result)

    @run_1k_times
    def test__depth(self):
        depth = rand_int()
        path_fct = faker.RandPath(depth=depth)
        result = path_fct()

        self.assertEqual(len(result.split("/")), depth + 1)

    @run_1k_times
    def test__root(self):
        root = rand_str()
        path_fct = faker.RandPath(root=root)
        result = path_fct()

        self.assertTrue(result.startswith(root))

    @run_1k_times
    def test__depth_root(self):
        root = rand_str()
        depth = rand_int()
        path_fct = faker.RandPath(depth=depth, root=root)
        result = path_fct()

        self.assertEqual(len(result.split("/")), depth + 1 + 1)
        self.assertTrue(result.startswith(root))


class RandFilePathTestCase(TestCase):
    @run_1k_times
    def test__default(self):
        filepath_fct = faker.RandFilePath()
        result = filepath_fct()

        self.assertFalse(result.endswith("/"))

        path, file = result.rsplit("/", 1)
        filename, ext = file.split(".", 1)

        # meh, I'm tired
        self.assertTrue(bool(len(path)))
        self.assertTrue(bool(len(file)))
        self.assertTrue(bool(len(filename)))
        self.assertTrue(bool(len(ext)))



from django.db import models

from django_utk.db.models.meta import ABCModelMeta

__all__ = [
    "ABCModel",
    "ModelMixin",
]


class ABCModel(models.Model, metaclass=ABCModelMeta):
    class Meta:
        abstract = True


class ModelMixin(models.Model):
    class Meta:
        abstract = True



from unittest import TestCase

from django_utk.tests import faker
from django_utk.tests.utils import run_1k_times

CHOICES = [42, 42.42, 42j, "foo", True, None, type, Exception, (), [], {}, ...]


class RandChoicesTestCase(TestCase):
    @run_1k_times
    def assertResultInChoices(self, choices_factory: faker.RandChoices, choices: list):
        for result in choices_factory():
            self.assertIn(result, choices)

    def test__default(self):
        for k in range(1, len(CHOICES) + len(CHOICES)):
            choices_fct = faker.RandChoices(CHOICES, k)
            self.assertResultInChoices(choices_fct, CHOICES)

    def test__no_choices(self):
        with self.assertRaises(IndexError) as _:
            choices_fct = faker.RandChoices([])
            choices_fct()


class RandChoiceTestCase(TestCase):
    def test__returns_one(self):
        choices_fct = faker.RandChoice(CHOICES)
        self.assertIn(choices_fct(), CHOICES)



from django_utk.db.models.softdelete.managers import SoftDeleteManager
from django_utk.db.models.softdelete.models import *
from django_utk.db.models.softdelete.models import __all__ as models_all
from django_utk.db.models.softdelete.querysets import *
from django_utk.db.models.softdelete.querysets import __all__ as querysets_all

__all__ = [
    *querysets_all,
    *models_all,
]

__all__ += [
    "SoftDeleteManager",
]



from abc import ABC, abstractmethod
from typing import Any, Mapping

from django.db import models
from django.db.models import Field

from django_utk.db.models.utils import get_model_meta as meta


class ModelFieldsTestCase(ABC):

    @property
    @abstractmethod
    def model(self) -> models.Model:
        raise NotImplemented

    @property
    @abstractmethod
    def required_fields(self) -> Mapping[str, Mapping[str, Any]]:
        raise NotImplemented

    def get_required_fields(self):
        return self.required_fields

    def test__fields(self):
        """
        Checking child-model's fields to make sure all parent's features in there. TODO: rewrite
        """
        required_fields = self.get_required_fields()

        model_meta = meta(self.model)
        model_fields = {field.attname: field for field in model_meta.fields}

        model_fields_names = set(model_fields.keys())
        required_fields_names = set(required_fields.keys())

        lost_fields_name = required_fields_names - model_fields_names

        self.assertFalse(
            lost_fields_name,
            f"Model does not implementing some of required fields: {lost_fields_name}",
        )

        for required_field_name, required_field_kw in required_fields.items():
            field: Field = model_fields[required_field_name]

            for kw_name, kw_value in required_field_kw.items():
                kw = getattr(field, kw_name)
                self.assertEqual(kw, kw_value)



import random

from django_utk.tests.faker.base import DataFactory
from django_utk.utils.typehint import typehint

__all__ = [
    "RandInt",
    "RandFloat",
]


class RandInt(DataFactory):
    getter = random.randint
    MIN = -2147483648  # PostgreSQL (4 bytes integer)
    MAX = +2147483647  # PostgreSQL (4 bytes integer)

    def __init__(self, a: int = MIN, b: int = MAX):
        super(RandInt, self).__init__(a=a, b=b)

    @typehint(DataFactory)
    def __call__(self, *, a: int = None, b: int = None) -> int:
        pass


class RandFloat(DataFactory):
    getter = random.uniform
    MIN = -1e300
    MAX = +1e300

    def __init__(self, a: float = MIN, b: float = MAX):
        super(RandFloat, self).__init__(a=a, b=b)

    @typehint(DataFactory)
    def __call__(self, *, a: float = None, b: float = None) -> float:
        pass
